import { TimeInSeconds } from "../midi/types";

declare var window: any;

function log10(x: number) {
  return Math.log(x) / Math.LN10;
}

export interface IAndySoundNode {
  readonly node?: AudioNode;
  readonly nodeInput?: AudioNode;
}

/*
  +-------------------+
  | IAndySoundNode    |
  | i - nodeInput     |
  | o - node          |
  +-------------------+

  +---------------------+
  | Sound               |
  | i - source (buffer) |
  | - volume (gain)     |
  | - panner (panner)   |
  | o - panner          |
  +---------------------+

  +--------------------+
  | SoundSystem        |
  | i - masterGainNode |
  | - analyser         |
  | o - speakers       |
  +--------------------+
  | play(Sound)        |
  +--------------------+
*/

/**
 * Sound represents a single sound file. For example an mp3, or ogg.
 * (Use the loadSound function to fetch a file). After you have a 
 * Sound you can use the play() method on the SoundSystem to play
 * the sound.
 */
export class Sound implements IAndySoundNode {
  public buffer: AudioBuffer;

  public loop: boolean;
  /** if looping where to start the loop (in seconds) */
  public loopStart: number;
  /** if looping where to end the loop (in seconds) */
  public loopEnd: number;

  public isInContext = false;
  // True if the sound is currently playing
  public isPlaying = false;
  // public isMusic = false;

  constructor(buffer: AudioBuffer) {
    this.buffer = buffer;
    this.loop = false;
    this.loopStart = 0;
    this.loopEnd = this.buffer.length;
  }

  private source: AudioBufferSourceNode | undefined;
  public volume: GainNode | undefined;
  public panner: PannerNode | undefined;

  setContext(context: AudioContext) {
    this.volume = context.createGain();
    // Note: we start with our volume all the way down
    // you need to set it to something by default
    this.volume.gain.setValueAtTime(0.0000001, 0);

    // TODO: the panner should be on a "track"
    this.panner = context.createPanner();
    this.panner.panningModel = 'HRTF';
    this.panner.distanceModel = 'inverse';
    this.panner.refDistance = 1;
    this.panner.maxDistance = 1000;
    this.panner.rolloffFactor = 1;
    this.panner.coneInnerAngle = 360;
    this.panner.coneOuterAngle = 0;
    this.panner.coneOuterGain = 0;
    // orientation looking down the z axis
    this.panner.orientationX.setValueAtTime(0, 0);
    this.panner.orientationY.setValueAtTime(0, 0);
    this.panner.orientationZ.setValueAtTime(-1, 0);
    this.panner.positionX.setValueAtTime(0, 0);
    this.panner.positionY.setValueAtTime(0, 0);
    this.panner.positionZ.setValueAtTime(0, 0);
    //////

    this.source = context.createBufferSource();

    this.source.onended = () => {
      this.isPlaying = false;
    }

    this.source.connect(this.volume);
    this.volume.connect(this.panner);

    this.source.buffer = this.buffer;
    this.source.loop = this.loop;
    this.isInContext = true;
  }

  get nodeInput(): AudioNode | undefined {
    return this.source;
  }

  get node(): AudioNode | undefined {
    // return (this.isMusic) ? this.volume : this.panner;
    // return this.volume;
    return this.panner;
  }

  play(time: number = 0) {
    if (!this.isInContext) {
      throw new Error('Sound not in context');
    }

    if (this.source && this.loop) {
      this.source.loop = this.loop;
      this.source.loopStart = this.loopStart;
      this.source.loopEnd = this.loopEnd;
      this.source.start(time);
    } else if (this.source) {
      this.source.loop = false;
      this.source.start(time);
    }
  }

  stop() {
    if (this.source)
      this.source.stop();
  }

  destroy() {
    this.source?.stop();
    if (this.volume)
      this.source?.disconnect(this.volume);
    if (this.panner)
      this.volume?.disconnect(this.panner);
  }
}

/**
 * This is the main crux of Andy. It sets up a bunch of default audio
 * nodes so you can control volume and panning and such and you can 
 * send it Sound files via play() to play sounds.
 */
export class SoundSystem implements IAndySoundNode {
  static addSoundToContext(context: AudioContext, sound: Sound) {
    sound.setContext(context);
  }
  static soundContext: AudioContext;
  static masterGainNode: GainNode;
  static analyser: AnalyserNode;

  constructor() {
    this.reboot();
  }

  context() {
    return SoundSystem.soundContext;
  }

  masterGainNode() {
    return SoundSystem.masterGainNode;
  }

  get nodeInput(): AudioNode | undefined {
    return this.masterGainNode();
  }

  reboot() {
    try {
      const AudioContext = window.webkitAudioContext // Safari and old versions of Chrome
        || window.AudioContext                       // Default
        || false;

      if (AudioContext) {
        if (SoundSystem.soundContext) {
          SoundSystem.soundContext.suspend();
        }
        SoundSystem.soundContext = new AudioContext();

        SoundSystem.soundContext.onstatechange = this.handleStateChange;

        SoundSystem.soundContext.audioWorklet.addModule('noise.processor.js');

        SoundSystem.analyser = SoundSystem.soundContext.createAnalyser();

        SoundSystem.masterGainNode = SoundSystem.soundContext.createGain();
        SoundSystem.masterGainNode.gain.value = 1;

        SoundSystem.masterGainNode.connect(SoundSystem.soundContext.destination);

        SoundSystem.masterGainNode.connect(SoundSystem.analyser);

        window.addEventListener('focus', (event: any) => this.resumePlay());
      } else {
        console.error('Andy: Web Audio API is not supported in this browser');
      }
    } catch (e) {
      console.error('Andy: Error creating Web Audio context', e);
    }
  }

  public getAnalyser(): AnalyserNode {
    return SoundSystem.analyser;
  }

  /** Get the current volume setting: 0 - 1 */
  public volume(): number {
    return SoundSystem.masterGainNode.gain.value;
  }

  /**
   * Gets the current volume setting in decibels
   */
  public volumeInDB(): number {
    return 20 * log10(this.volume());
  }

  /**
   * Sets the current volume using decibels
   */
  public setVolumeInDB(decibel: number) {
    this.adjustVolume(Math.pow(10, (decibel / 20)));
  }

  /** Set the current volume to: 0 - 1 */
  public adjustVolume(by: number) {
    SoundSystem.masterGainNode.gain.value = by;
  }

  /** Play the given sound in the system. Note the sound does not
   * have to be connected to anything, this will connect the sound 
   * to the master gain node before playing.
   * 
   * Time is an offset within the sound to play from. e.g. 0 is from
   * the start of the sample.
   */
  public play(sound: Sound, multiple = false, time: TimeInSeconds = 0) {
    if (multiple === true || !sound.isPlaying) {
      this.playSound(sound, time);
    }
  }

  public stop(sound: Sound) {
    if (sound.isPlaying) {
      sound.stop();
      sound.isPlaying = false;
    }
  }

  /** Get the sound contexts precise current time 
   * (ever-increasing hardware timestamp in seconds) */
  public currentTime(): TimeInSeconds {
    if (!SoundSystem.soundContext) {
      return 0;
    }
    return SoundSystem.soundContext.currentTime;
  }

  /**
   * Given an array buffer of wav, mp3, ogg, etc try to turn it
   * into an audio buffer we can play. 
   */
  public decodeSound(data: ArrayBuffer): Promise<AudioBuffer> {
    return new Promise<AudioBuffer>((res, rej) => {
      let buffer: AudioBuffer;
      if (SoundSystem.soundContext) {
        SoundSystem.soundContext.decodeAudioData(data, (ab: AudioBuffer) => {
          buffer = ab;
          res(buffer);
        },
          (error: any) => {
            console.error(error);
            rej(error);
          }
        );
      }
    });
  }

  public resumePlay(): Promise<void> {
    return SoundSystem.soundContext.resume();
  }

  private playSound(sound: Sound, time: TimeInSeconds = 0) {
    // Put the sound into our current context
    sound.setContext(this.context());
    // connect to current master gain node
    if (sound.node) {
      sound.node.connect(this.masterGainNode());
    }
    // default
    sound.volume?.gain.setValueAtTime(0.8, time);
    sound.play(time);
    sound.isPlaying = true;
  }

  private handleStateChange(e: Event): any {
    if (SoundSystem.soundContext && SoundSystem.soundContext.state !== 'running') {
      this.resumePlay();
    }
    return undefined;
  }
}

/**
 * Load a sound from the server, and pass it through the 
 * sound system to try to decode the file into mp3, etc
 */
export const loadSound = (url: string, soundSystem: SoundSystem): Promise<Sound> => {
  return new Promise((resolve, reject) => {
    const headers = new Headers();
    headers.append('User-Agent', 'Andy/1.0');
    // if (mimeType) {
    //   headers.append('Accept', mimeType);
    //   headers.append('Content-Type', mimeType);
    // }
    const body = undefined;
    fetch(url, {
      method: 'GET',
      mode: 'cors',
      cache: 'reload',
      headers: headers,
      credentials: 'same-origin',
      redirect: 'follow',
      referrerPolicy: 'no-referrer',
      body: (body) ? JSON.stringify(body) : body,
    })
      .then(response => response.arrayBuffer())
      .then(response => {
        soundSystem.decodeSound(<ArrayBuffer>response).then(
          (b: AudioBuffer) => {
            const sound = new Sound(b);
            resolve(sound);
          },
          (err) => {
            reject(err);
          }).catch(e => console.error(e));
      }).catch(e => console.error(e));
  });
};
