
export class BinaryFile {
  dataView: DataView;
  buffer: ArrayBuffer;
  offset: number;

  constructor(data: ArrayBuffer) {
    this.buffer = data;
    this.dataView = new DataView(this.buffer, 0);
    this.offset = 0;
  }

  tell(): number {
    return this.offset;
  }

  seek(offset: number) {
    this.offset = offset;
  }

  skip(bytes: number) {
    this.offset += bytes;
  }

  eof() {
    return this.offset >= this.buffer.byteLength; // length;
  }

  slice(offset: number, length: number) {
    return new BinaryFile(this.buffer.slice(offset, offset + length));
  }

  read(length: number): BinaryFile {
    const result = new BinaryFile(this.buffer.slice(this.offset, this.offset + length));
    this.offset += length;
    return result;
  }

  readString(length: number): string {
    let result = '';
    let terminated = false;
    for (let i = 0; i < length; i++) {
      const byte = this.readUInt8();
      if (byte === 0x0) {
        terminated = true;
      }
      if (!terminated) {
        result += String.fromCharCode(byte);
      }
    }
    return result;
  }

  readCString() {
    let result = '';
    for (let i = 0; i < 128; i++) {
      const byte = this.readUInt8();
      if (byte === 0x0) {
        return result;
      }
      result += String.fromCharCode(byte);
    }
    return result;
  }

  readChars(count: number): Uint8Array {
    return this.readBytes(count);
  }

  readBytes(count: number): Uint8Array {
    let result = new Uint8Array(count);
    for (let i = 0; i < count; i++) {
      const byte = this.readUInt8();
      result[i] = byte;
    }
    return result;
  }

  readSBytes(count: number): Int8Array {
    let result = new Int8Array(count);
    for (let i = 0; i < count; i++) {
      const byte = this.readInt8();
      result[i] = byte;
    }
    return result;
  }

  readUInt8(): number {
    return this.dataView.getUint8(this.offset++);
  }

  readInt8() {
    return this.dataView.getInt8(this.offset++);
  }

  readUInt16(little = true) {
    const result = this.dataView.getUint16(this.offset, little);
    this.offset += 2;
    return result;
  }

  readInt16(little = true) {
    const result = this.dataView.getInt16(this.offset, little);
    this.offset += 2;
    return result;
  }

  readUInt32(little = true) {
    const result = this.dataView.getUint32(this.offset, little);
    this.offset += 4;
    return result;
  }

  readInt32(little = true) {
    const result = this.dataView.getInt32(this.offset, little);
    this.offset += 4;
    return result;
  }

  readFloat(little = true) {
    const result = this.dataView.getFloat32(this.offset, little);
    this.offset += 4;
    return result;
  }

  close() {
    // this.dataView = undefined;
    // this.buffer = undefined;
    this.offset = -1;
  }
}
