import { SoundSystem } from "../digital/audio";
import { Format, MIDI } from "./song";
import { GeneralMidiInstruments, GeneralMIDIPercussion, noteByMidi } from "../midi/types";
import { createPlayHeadClock, toSignature, createTempo, PerTickCallback, TickCallbackContainer } from "../metronome";
import { patchPlayer } from "./audio";
import { AndyPatch, AndyPercussionPatch } from './patchBank';
import { SoundOn, SoundOff, NoOp, Tracker, Track } from "../digital/types";

const MAX_CHANNELS = 16;

type MidiSong = {
  /** the "raw" midi song (with events, etc) */
  song: MIDI;
}

export type MidiWithTracks = MidiSong & Tracker;

export type MidiRun = MidiWithTracks & TickCallbackContainer;

export function createRun(system: SoundSystem, song: MIDI): MidiRun {
  if (song.format === Format.MultiAsync)
    throw new Error("Currently only support synchronous tracks");

  const run: MidiRun = {
    song: song,
    tracks: [],
    output: system.context().createGain(),
    tickListeners: []
  };

  for (let i = 0; i < song.numberOfTracks; i++) {
    const pan = system.context().createPanner();
    pan.orientationX.setValueAtTime(0, 0);
    pan.orientationY.setValueAtTime(0, 0);
    pan.orientationZ.setValueAtTime(-1, 0);
    pan.positionX.setValueAtTime(0, 0);
    pan.positionY.setValueAtTime(1, 0);
    pan.positionZ.setValueAtTime(0, 0);

    const vol = system.context().createGain();
    pan.connect(vol);
    vol.connect(run.output);

    vol.gain.setValueAtTime(1, 0);

    const startFns = new Array<Array<SoundOff>>(MAX_CHANNELS);
    for (let i = 0; i < MAX_CHANNELS; i++) {
      startFns[i] = []; // NoOp;
    }

    const t = {
      lastFn: startFns,
      idx: 0,
      vol,
      pan,
    } as Track;

    run.tracks.push(t);
  }

  // connect the song output to the speakers
  run.output.connect(system.masterGainNode());
  return run;
}

/** Clean up after the song run has played */
export function destroyRun(system: SoundSystem, run: MidiRun) {
  // TODO: disconnect tracks?
  run.output.disconnect(system.masterGainNode());
  while (run.tickListeners.length) {
    run.tickListeners.pop();
  }
}

const useInstrument = (system: SoundSystem, run: MidiRun, patchNum: number): SoundOn => {
  const patch = AndyPatch.get(patchNum);
  if (patch)
    return patchPlayer(system, patch);
  else {
    const p = AndyPatch.get(GeneralMidiInstruments.ChurchOrgan);
    if (p)
      return patchPlayer(system, p);
  }

  return (NoOp as any) as SoundOn;
};

const usePercussion = (system: SoundSystem, run: MidiRun, patchNum: number): SoundOn => {
  const patch = AndyPercussionPatch.get(patchNum);
  if (patch)
    return patchPlayer(system, patch);

  return (NoOp as any) as SoundOn;
};

export function silenceRun(songRun: MidiRun) {
  songRun.tracks.forEach(track => {
    for (let i = 0; i < MAX_CHANNELS; i++) {
      const fns = track.lastFn[i];
      while (fns.length) {
        const soundOff = fns.pop();
        if (soundOff) soundOff();
      }
    }
  });
}

export function setRunTimeTo(time: number, songRun: MidiRun) {
  const len = songRun.song.tracks.length;

  for (let t = 0; t < len; t++) {
    const elen = songRun.song.tracks[t].events.length;
    // Set the track index to the index closest to the time
    // we are asked to.
    for (let e = 0; e < elen; e++) {
      const abstime = songRun.song.tracks[t].events[e].abstime;
      if (abstime >= time) {
        songRun.tracks[t].idx = e;
        break;
      }
    }
  }
}

export function rewindRun(songRun: MidiRun) {
  songRun.tracks.forEach(track => {
    track.idx = 0;
  });
}

// TODO: this needs, not only a refactor, but for V2 this should go back
// to sending out raw midi events so the whole thing needs
function handleEventCallback(system: SoundSystem, run: MidiRun): PerTickCallback {
  let tempo = createTempo(run.song.ticksPerQuarterNote, 500000);
  let sig = toSignature("4/4");

  const tickCallBack: PerTickCallback = (tick, time, bi) => {
    const trackLen = run.song.tracks.length

    const stopAllSounds = (track: Track) => {
      for (let i = 0; i < MAX_CHANNELS; i++) {
        const fns = track.lastFn[i];
        while (fns.length) {
          const soundOff = fns.pop();
          if (soundOff) soundOff();
        }
      }
    }

    // go through all tracks...
    for (let t = 0; t < trackLen; t++) {
      const track = run.tracks[t];
      const events = run.song.tracks[t].events;

      // go through all events in the track, starting from the last
      // index we saw
      for (let e = track.idx; e < events.length; e++) {
        const ce = events[e];

        // if the abstime is not this tick, then get out of this
        // loop and set it to check next time
        if (ce.abstime !== tick) {
          track.idx = e;
          break;
        }

        if (ce.type === "Meta") {
          if (ce.subtype === "SetTempo") {
            // TODO: immutable?
            tempo = createTempo(tempo.ppq, ce.data[0] as number);
          }

          if (ce.subtype === "TimeSignature") {
            sig.noteBeats = ce.data[0] as number;
            sig.perMeasure = ce.data[1] as number;
            sig.ticksInMetronome = ce.data[2] as number;
            sig.toQuarter32nd = ce.data[3] as number;
          }
        }

        if (ce.type === "Channel") {
          if (ce.subtype === "ProgramChange") {
            track.instrument = useInstrument(system, run, ce.data[0] as number);
          }

          if (ce.subtype === "Controller") {
            // Volume
            if (ce.data[0] === 0x7) {
              track.vol.gain.setValueAtTime(
                (ce.data[1] as number / 127),
                time
              );
            }
            //Pan
            if (ce.data[0] === 0xa) {
              const amt = (ce.data[1] as number - 127 + 63.5) / 63.5;
              // Pan is really sensitive so dumb it down a lot.
              const amtScaled = Math.round(amt * 10) / 10;
              track.pan.positionX.setValueAtTime(amtScaled, time);
            }
          }

          if (ce.subtype === "NoteOff") {
            const fns = track.lastFn[ce.channel ?? 0];
            while (fns.length) {
              const soundOff = fns.pop();
              if (soundOff) soundOff();
            }
          }

          if (ce.subtype === "NoteOn") {
            // @ts-ignore
            const mnote = ce.data[0] ?? 60;
            const note = noteByMidi(mnote as number);

            // Drums
            if (ce.channel === 0x9) {
              track.lastFn[ce.channel].push(
                usePercussion(system, run, mnote as number)
                  (note, system.currentTime(), track.pan)
              );
            } else {
              if (track.instrument) {
                // lame version of velocity
                track.vol.gain.setValueAtTime(
                  (ce.data[1] as number / 127), system.currentTime()
                );
                track.lastFn[ce.channel ?? 0].push(
                  track.instrument(note, system.currentTime(), track.pan)
                );
              } else {
                console.warn(`NoteOn event, but no instrument selected. Channel ${ce.channel ?? 0} Track ${t} Event ${e}`);
              }
            }
          }

          if (ce.subtype === "EndOfTrack") {
            stopAllSounds(track);
          }
        }
      }
    }

    return [tempo, sig]
  }

  return tickCallBack;
}

/**
 * Sets up the default tempo and signature, and also starts
 * the metronome timer for use in playback and events. Once
 * this is done, you should call `playHeadToggle()` to start
 * the actual song playing.
 */
export function insertRun(system: SoundSystem, run: MidiRun) {
  // default 120bpm
  let tempo = createTempo(run.song.ticksPerQuarterNote, 500000);
  let sig = toSignature("4/4");

  run.tickListeners.push(handleEventCallback(system, run));

  createPlayHeadClock(system, tempo, sig, run);
}