import { BinaryFile } from "../io/file";

/*
  https://www.personal.kent.edu/~sbirch/Music_Production/MP-II/MIDI/midi_file_format.htm
  https://faydoc.tripod.com/formats/mid.htm
  http://www.petesqbsite.com/sections/express/issue18/midifilespart1.html
*/

enum ChunkType {
  Header = "MThd",
  Track = "MTrk",
}

/** 
 * Single track is fairly self-explanatory - one track only. 
 * Synchronous multiple tracks means that the tracks will all be vertically synchronous, 
 * or in other words, they all start at the same time, and so can represent different 
 * parts in one song. Asynchronous multiple tracks do not necessarily start at the 
 * same time, and can be completely asynchronous. 
 */
export enum Format {
  Single = 0,      // single-track
  MultiSync = 1,   // 1 - multiple tracks, synchronous
  /** _WARNING_: Unsupported in Andy */
  MultiAsync = 2,  // 2 - multiple tracks, asynchronous
}

type MetaSubtype = 'NoteOn' | 'NoteOff' | 'NoteAfterTouch' | 'Controller' | 'ProgramChange' | 'ChannelAfterTouch' | 'PitchBend';

type EventSubtype = 'SequenceNumber' | 'Text' | 'CopyrightNotice' | 'TrackName' | 'InstrumentName'
  | 'Lyrics' | 'Marker' | 'CuePoint' | 'MidiChannelPrefix' | 'EndOfTrack' | 'SetTempo'
  | 'SMPTEOffset' | 'TimeSignature' | 'KeySignature' | 'SequencerSpecific' | 'Unknown'
  | MetaSubtype;

type MidiEvent = {
  /** The raw midi event type */
  raw: number,
  /** The delta time from the last event */
  time: number,
  /** The time the event occurs in the song */
  abstime: number,
  type: 'Meta' | 'SysEx' | 'Channel',
  channel?: number,
  subtype?: EventSubtype,
  /** The raw midi event data. `raw` plus this will recreate the 
   * original midi message */
  data: Array<number | string | Uint8Array>,
};

type Track = {
  name: string,
  events: Array<MidiEvent>;
}

export type Marker = {
  /** The delta time from the last event */
  time: number,
  /** The time the event occurs in the song */
  abstime: number,
}

export class MIDI {
  private data: BinaryFile;

  /** Ticks per quarter note. aka PPQ */
  public ticksPerQuarterNote: number;
  public format: Format;
  /** number of tracks */
  public numberOfTracks: number;
  /** Tracks and midi events */
  public tracks: Array<Track>;

  public markers: Map<string, Marker> = new Map([]);

  constructor(data: ArrayBuffer) {
    this.data = new BinaryFile(data);
    this.format = 0;
    this.numberOfTracks = 0;
    this.ticksPerQuarterNote = 0;

    this.tracks = new Array<Track>(this.numberOfTracks);

    this.readHeader();
    for (let t = 0; t < this.numberOfTracks; t++) {
      this.tracks[t] = this.readTrack();
    }
  }

  private readHeader() {
    // 4 + 4 + 2 + 2 + 2 = 14
    const hdr = this.data.readString(4);
    const size = this.data.readUInt32(false);
    if (hdr !== ChunkType.Header || size !== 6) {
      throw new Error(`Incorrect header, not a midi file? ${hdr} ${size}`);
    }

    const format = this.data.readUInt16(false);
    this.format = format;

    const tracks = this.data.readUInt16(false);
    this.numberOfTracks = tracks;

    const ticks = this.data.readUInt16(false);
    if ((ticks & 0x8000) === 0) {
      this.ticksPerQuarterNote = ticks;
    } else {
      throw new Error(
        "Divisions set as units per SMTPE frame: Unsupported"
      );
    }
  }

  private readTrack(): Track {
    const events: MidiEvent[] = [];
    const track: Track = {
      name: "Unknown",
      events,
    };

    const hdr = this.data.readString(4);
    // The 4 bytes after MTrk give the length of the track in bytes. 
    const length = this.data.readUInt32(false);
    const start = this.data.tell();

    // a delta-time is the duration (in clocks) between an event and the preceding event.
    // The largest value allowed within a MIDI file is `0F FF FF FF`. This limit is set to 
    // allow variable-length quantities to be manipulated as 32-bit integers. 
    // TODO: not right...
    let lastEvent = 0;
    let trackTime = 0;

    while (true) {
      let dtime = this.readVariableLen();
      let eventType = this.data.readUInt8();
      trackTime += dtime;

      if (this.data.tell() >= (start + length)) {
        break;
      }

      if ((eventType & 0xF0) === 0xF0) {
        if (eventType === 0xFF) {
          const event = this.metaEvents(dtime, trackTime, eventType, track);
          events.push(event);
          if (event.raw === 0x2F) {
            // End of track, break out of the event loop
            break;
          }
        } else if (eventType === 0xF0 || eventType === 0xF7) {
          const event = this.sysexEvents(dtime, trackTime, eventType);
          events.push(event);
        } else {
          console.warn(`Unknown event ${eventType.toString(16)}`);
        }
      } else {
        const event = this.midiEvents(dtime, trackTime, eventType, lastEvent);
        lastEvent = event.raw;
        events.push(event);
      }
    }

    return track;
  }

  /** Returns true on end of track, false if you should keep reading */
  private metaEvents(dtime: number, ttime: number, eventType: number, track: Track): MidiEvent {
    const event: MidiEvent = {
      raw: eventType,
      time: dtime,
      abstime: ttime,
      type: 'Meta',
      data: [],
    };

    // The possible range is 00-7F. Not all values in this range are defined, 
    // but programs must be able to cope with (ie ignore) unexpected values by 
    // examining the length and skipping over the data portion. 
    const command = this.data.readUInt8();
    // number of bytes that follow
    const length = this.readVariableLen();

    event.raw = command;
    switch (command) {
      case 0x00: {
        event.subtype = 'SequenceNumber';
        event.data?.push(this.data.readUInt16(false));
      } break;
      case 0x01: {
        event.subtype = 'Text';
        const txt = this.data.readString(length);
        event.data?.push(txt);
      } break;
      case 0x02: {
        event.subtype = 'CopyrightNotice';
        const copy = this.data.readString(length);
        event.data?.push(copy);
      } break;
      case 0x03: {
        event.subtype = 'TrackName';
        const name = this.data.readString(length);
        event.data?.push(name);
        track.name = name;
      } break;
      case 0x04: {
        event.subtype = 'InstrumentName';
        const name = this.data.readString(length);
        event.data?.push(name);
      } break;
      case 0x05: {
        event.subtype = 'Lyrics';
        const lyrics = this.data.readString(length);
        event.data?.push(lyrics);
      } break;
      case 0x06: {
        event.subtype = 'Marker';
        const markerStr = this.data.readString(length);
        event.data?.push(markerStr);
        this.markers.set(markerStr, { time: dtime, abstime: ttime });
      } break;
      case 0x07: {
        event.subtype = 'CuePoint';
        const cueStr = this.data.readString(length);
        event.data?.push(cueStr);
      } break;
      case 0x0f: {
        // Reaper MIDI export can have this set for some reason
        // Data says something like: NOTE 0 43 staff 3
        event.subtype = 'Unknown';
        if (length > 0) {
          const str = this.data.readBytes(length);
          event.data?.push(str);
        }
      } break;
      case 0x20: {
        event.subtype = 'MidiChannelPrefix';
        event.data?.push(this.data.readUInt8());
      } break;
      case 0x21: {
        // ff 21 01 vv => prefix port 
        event.subtype = 'Unknown';
        event.data?.push(this.data.readUInt8());
      } break;
      case 0x2f: {
        event.subtype = 'EndOfTrack';
      } break;
      case 0x51: {
        // Set tempo
        event.subtype = 'SetTempo';
        const microseconds =
          (this.data.readUInt8() << 16) +
          (this.data.readUInt8() << 8) +
          this.data.readUInt8();
        event.data?.push(microseconds);
      } break;
      case 0x54: {
        event.subtype = 'SMPTEOffset';
        if (length > 0)
          this.data.readBytes(length);
        // not supported.
      } break;
      case 0x58: {
        // Time Signature
        event.subtype = 'TimeSignature';
        event.data?.push(this.data.readUInt8());
        event.data?.push(Math.pow(2, this.data.readUInt8()))
        event.data?.push(this.data.readUInt8());
        event.data?.push(this.data.readUInt8());
      } break;
      case 0x59: {
        // Key signature
        event.subtype = 'KeySignature';
        // key / scale
        event.data?.push(this.data.readInt8());
        event.data?.push(this.data.readUInt8());
      } break;
      case 0x7F: {
        event.subtype = "SequencerSpecific";
        const _data = this.data.readBytes(length);
      } break;
      default: {
        console.warn(`Unknown meta event ${eventType.toString(16)} ${command.toString(16)} ${length}`);
        event.subtype = 'Unknown';
        if (length > 0)
          event.data?.push(this.data.readBytes(length));
      }
    }

    return event;
  }

  private sysexEvents(dtime: number, ttime: number, eventType: number): MidiEvent {
    const event: MidiEvent = {
      raw: eventType,
      abstime: ttime,
      time: dtime,
      type: 'SysEx',
      data: [],
    };

    const length = this.readVariableLen();
    const _data = this.data.readBytes(length);
    return event;
  }

  private midiEvents(dtime: number, ttime: number, command: number, lastEvent: number): MidiEvent {
    let useEvent = 0;
    let param = 0;

    // This is confusing... we need to keep track
    // of the last command, and if the current command &0x80 is not
    // set, then we need to use the last command. This is used to
    // save space in the midi file :-/
    if ((command & 0x80) === 0) {
      param = command;
      useEvent = lastEvent;
    } else {
      param = this.data.readUInt8();
      useEvent = command;
    }

    const event: MidiEvent = {
      raw: useEvent,
      time: dtime,
      abstime: ttime,
      type: 'Channel',
      data: [],
    };

    const cmd = ((useEvent & 0xF0) >> 4);
    const channel = ((useEvent & 0x0F));
    event.channel = channel;

    switch (cmd) {
      case 0x8: {
        event.subtype = 'NoteOff';
        const note = param;
        event.data?.push(note);
        const velocity = this.data.readUInt8();
        event.data?.push(velocity);
      } break;
      case 0x9: {
        const note = param;
        event.data?.push(note);
        const velocity = this.data.readUInt8();
        event.data?.push(velocity);
        if (velocity == 0)
          event.subtype = 'NoteOff';
        else
          event.subtype = 'NoteOn';
      } break;
      case 0xa: {
        event.subtype = 'NoteAfterTouch';
        event.data?.push(param);
        const amount = this.data.readUInt8();
        event.data?.push(amount);
      } break;
      case 0xb: {
        event.subtype = 'Controller';
        // controller number
        event.data?.push(param);
        // new value
        const value = this.data.readUInt8();
        event.data?.push(value);
      } break;
      case 0xc: {
        event.subtype = 'ProgramChange';
        // new program (patch) number
        const type = param;
        event.data?.push(type);
      } break;
      case 0xd: {
        event.subtype = 'ChannelAfterTouch';
        event.data?.push(param);
      } break;
      case 0xe: {
        event.subtype = 'PitchBend';
        // Pitch wheel change (2000H is normal or no change)
        // bb - bottom (least sig) 7 bits of value
        // tt - top (most sig) 7 bits of value
        const amount = param + (this.data.readUInt8() << 7);
        event.data?.push(amount);
      } break;
      default:
        console.warn(`[[UNKNOWN ${command.toString(16)}]]`);
    }

    return event;
  }

  /**
   * Reads a variable length variable assumes the data read
   * head is all set.
   */
  private readVariableLen(): number {
    const start = this.data.readUInt8();
    if ((start & 0x80) === 0) {
      return start;
    }

    let c = 0;
    let value = start & 0x7F;
    do {
      value = (value << 7) + ((c = this.data.readUInt8() & 0x7F));
    } while (c & 0);

    return value;
  }
}


/**
 * Load a midi from the server, and pass it through the 
 * sound system to try to setup the file for play
 */
export const loadMidi = (url: string): Promise<MIDI> => {
  return new Promise((resolve, reject) => {
    const headers = new Headers();
    headers.append('User-Agent', 'Andy/1.0');
    headers.append('Accept', 'audio/midi');
    const body = undefined;
    fetch(url, {
      method: 'GET',
      mode: 'cors',
      cache: 'reload',
      headers: headers,
      credentials: 'same-origin',
      redirect: 'follow',
      referrerPolicy: 'no-referrer',
      body: (body) ? JSON.stringify(body) : body,
    })
      .then(response => response.arrayBuffer())
      .then(response => {
        const mid = new MIDI(response);
        resolve(mid);
      }).catch(e => console.error(e));
  });
};

