(() => {
  // src/digital/audio.ts
  function log10(x) {
    return Math.log(x) / Math.LN10;
  }
  var Sound = class {
    constructor(buffer) {
      this.isInContext = false;
      this.isPlaying = false;
      this.buffer = buffer;
      this.loop = false;
      this.loopStart = 0;
      this.loopEnd = this.buffer.length;
    }
    setContext(context) {
      this.volume = context.createGain();
      this.volume.gain.setValueAtTime(1e-7, 0);
      this.panner = context.createPanner();
      this.panner.panningModel = "HRTF";
      this.panner.distanceModel = "inverse";
      this.panner.refDistance = 1;
      this.panner.maxDistance = 1e3;
      this.panner.rolloffFactor = 1;
      this.panner.coneInnerAngle = 360;
      this.panner.coneOuterAngle = 0;
      this.panner.coneOuterGain = 0;
      this.panner.orientationX.setValueAtTime(0, 0);
      this.panner.orientationY.setValueAtTime(0, 0);
      this.panner.orientationZ.setValueAtTime(-1, 0);
      this.panner.positionX.setValueAtTime(0, 0);
      this.panner.positionY.setValueAtTime(0, 0);
      this.panner.positionZ.setValueAtTime(0, 0);
      this.source = context.createBufferSource();
      this.source.onended = () => {
        this.isPlaying = false;
      };
      this.source.connect(this.volume);
      this.volume.connect(this.panner);
      this.source.buffer = this.buffer;
      this.source.loop = this.loop;
      this.isInContext = true;
    }
    get nodeInput() {
      return this.source;
    }
    get node() {
      return this.panner;
    }
    play(time = 0) {
      if (!this.isInContext) {
        throw new Error("Sound not in context");
      }
      if (this.source && this.loop) {
        this.source.loop = this.loop;
        this.source.loopStart = this.loopStart;
        this.source.loopEnd = this.loopEnd;
        this.source.start(time);
      } else if (this.source) {
        this.source.loop = false;
        this.source.start(time);
      }
    }
    stop() {
      if (this.source)
        this.source.stop();
    }
    destroy() {
      this.source?.stop();
      if (this.volume)
        this.source?.disconnect(this.volume);
      if (this.panner)
        this.volume?.disconnect(this.panner);
    }
  };
  var SoundSystem = class {
    static addSoundToContext(context, sound) {
      sound.setContext(context);
    }
    constructor() {
      this.reboot();
    }
    context() {
      return SoundSystem.soundContext;
    }
    masterGainNode() {
      return SoundSystem.masterGainNode;
    }
    get nodeInput() {
      return this.masterGainNode();
    }
    reboot() {
      try {
        const AudioContext = window.webkitAudioContext || window.AudioContext || false;
        if (AudioContext) {
          if (SoundSystem.soundContext) {
            SoundSystem.soundContext.suspend();
          }
          SoundSystem.soundContext = new AudioContext();
          SoundSystem.soundContext.onstatechange = this.handleStateChange;
          SoundSystem.soundContext.audioWorklet.addModule("noise.processor.js");
          SoundSystem.analyser = SoundSystem.soundContext.createAnalyser();
          SoundSystem.masterGainNode = SoundSystem.soundContext.createGain();
          SoundSystem.masterGainNode.gain.value = 1;
          SoundSystem.masterGainNode.connect(SoundSystem.soundContext.destination);
          SoundSystem.masterGainNode.connect(SoundSystem.analyser);
          window.addEventListener("focus", (event) => this.resumePlay());
        } else {
          console.error("Andy: Web Audio API is not supported in this browser");
        }
      } catch (e) {
        console.error("Andy: Error creating Web Audio context", e);
      }
    }
    getAnalyser() {
      return SoundSystem.analyser;
    }
    volume() {
      return SoundSystem.masterGainNode.gain.value;
    }
    volumeInDB() {
      return 20 * log10(this.volume());
    }
    setVolumeInDB(decibel) {
      this.adjustVolume(Math.pow(10, decibel / 20));
    }
    adjustVolume(by) {
      SoundSystem.masterGainNode.gain.value = by;
    }
    play(sound, multiple = false, time = 0) {
      if (multiple === true || !sound.isPlaying) {
        this.playSound(sound, time);
      }
    }
    stop(sound) {
      if (sound.isPlaying) {
        sound.stop();
        sound.isPlaying = false;
      }
    }
    currentTime() {
      if (!SoundSystem.soundContext) {
        return 0;
      }
      return SoundSystem.soundContext.currentTime;
    }
    decodeSound(data) {
      return new Promise((res, rej) => {
        let buffer;
        if (SoundSystem.soundContext) {
          SoundSystem.soundContext.decodeAudioData(data, (ab) => {
            buffer = ab;
            res(buffer);
          }, (error) => {
            console.error(error);
            rej(error);
          });
        }
      });
    }
    resumePlay() {
      return SoundSystem.soundContext.resume();
    }
    playSound(sound, time = 0) {
      sound.setContext(this.context());
      if (sound.node) {
        sound.node.connect(this.masterGainNode());
      }
      sound.volume?.gain.setValueAtTime(0.8, time);
      sound.play(time);
      sound.isPlaying = true;
    }
    handleStateChange(e) {
      if (SoundSystem.soundContext && SoundSystem.soundContext.state !== "running") {
        this.resumePlay();
      }
      return void 0;
    }
  };
  var loadSound = (url, soundSystem) => {
    return new Promise((resolve, reject) => {
      const headers = new Headers();
      headers.append("User-Agent", "Andy/1.0");
      const body = void 0;
      fetch(url, {
        method: "GET",
        mode: "cors",
        cache: "reload",
        headers,
        credentials: "same-origin",
        redirect: "follow",
        referrerPolicy: "no-referrer",
        body: body ? JSON.stringify(body) : body
      }).then((response) => response.arrayBuffer()).then((response) => {
        soundSystem.decodeSound(response).then((b) => {
          const sound = new Sound(b);
          resolve(sound);
        }, (err) => {
          reject(err);
        }).catch((e) => console.error(e));
      }).catch((e) => console.error(e));
    });
  };

  // src/digital/types.ts
  var NoOp = () => {
  };

  // src/midi/types.ts
  function noteByMidi(value) {
    const midiVals = [...NoteToMidi.entries()].filter(({1: v}) => v === value).map(([k]) => k);
    return midiVals[0];
  }
  var NoteToMidi = new Map([
    ["G-9", 127],
    ["Gb9", 126],
    ["F#9", 126],
    ["F-9", 125],
    ["E-9", 124],
    ["Eb9", 123],
    ["D#9", 123],
    ["D-9", 122],
    ["Db9", 121],
    ["C#9", 121],
    ["C-9", 120],
    ["B-9", 119],
    ["Bb8", 118],
    ["A#8", 118],
    ["A-8", 117],
    ["Ab8", 116],
    ["G#8", 116],
    ["G-8", 115],
    ["Gb8", 114],
    ["F#8", 114],
    ["F-8", 113],
    ["E-8", 112],
    ["Eb8", 111],
    ["D#8", 111],
    ["D-8", 110],
    ["Db8", 109],
    ["C#8", 109],
    ["C-8", 108],
    ["B-7", 107],
    ["Bb7", 106],
    ["A#7", 106],
    ["A-7", 105],
    ["Ab7", 104],
    ["G#7", 104],
    ["G-7", 103],
    ["Gb7", 102],
    ["F#7", 102],
    ["F-7", 101],
    ["E-7", 100],
    ["Eb7", 99],
    ["D#7", 99],
    ["D-7", 98],
    ["Db7", 97],
    ["C#7", 97],
    ["C-7", 96],
    ["B-6", 95],
    ["Bb6", 94],
    ["A#6", 94],
    ["A-6", 93],
    ["Ab6", 92],
    ["G#6", 92],
    ["G-6", 91],
    ["Gb6", 90],
    ["F#6", 90],
    ["F-6", 89],
    ["E-6", 88],
    ["Eb6", 87],
    ["D#6", 87],
    ["D-6", 86],
    ["Db6", 85],
    ["C#6", 85],
    ["C-6", 84],
    ["B-5", 83],
    ["Bb5", 82],
    ["A#5", 82],
    ["A-5", 81],
    ["Ab5", 80],
    ["G#5", 80],
    ["G-5", 79],
    ["Gb5", 78],
    ["F#5", 78],
    ["F-5", 77],
    ["E-5", 76],
    ["Eb5", 75],
    ["D#5", 75],
    ["D-5", 74],
    ["Db5", 73],
    ["C#5", 73],
    ["C-5", 72],
    ["B-4", 71],
    ["Bb4", 70],
    ["A#4", 70],
    ["A-4", 69],
    ["Ab4", 68],
    ["G#4", 68],
    ["G-4", 67],
    ["Gb4", 66],
    ["F#4", 66],
    ["F-4", 65],
    ["E-4", 64],
    ["Eb4", 63],
    ["D#4", 63],
    ["D-4", 62],
    ["Db4", 61],
    ["C#4", 61],
    ["C-4", 60],
    ["B-3", 59],
    ["Bb3", 58],
    ["A#3", 58],
    ["A-3", 57],
    ["Ab3", 56],
    ["G#3", 56],
    ["G-3", 55],
    ["Gb3", 54],
    ["F#3", 54],
    ["F-3", 53],
    ["E-3", 52],
    ["Eb3", 51],
    ["D#3", 51],
    ["D-3", 50],
    ["Db3", 49],
    ["C#3", 49],
    ["C-3", 48],
    ["B-2", 47],
    ["Bb2", 46],
    ["A#2", 46],
    ["A-2", 45],
    ["Ab2", 44],
    ["G#2", 44],
    ["G-2", 43],
    ["Gb2", 42],
    ["F#2", 42],
    ["F-2", 41],
    ["E-2", 40],
    ["Eb2", 39],
    ["D#2", 39],
    ["D-2", 38],
    ["Db2", 37],
    ["C#2", 37],
    ["C-2", 36],
    ["B-1", 35],
    ["Bb1", 34],
    ["A#1", 34],
    ["A-1", 33],
    ["Ab1", 32],
    ["G#1", 32],
    ["G-1", 31],
    ["Gb1", 30],
    ["F#1", 30],
    ["F-1", 29],
    ["E-1", 28],
    ["Eb1", 27],
    ["D#1", 27],
    ["D-1", 26],
    ["Db1", 25],
    ["C#1", 25],
    ["C-1", 24],
    ["B-0", 23],
    ["Bb0", 22],
    ["A#0", 22]
  ]);
  var MidiToFreq = [
    8.18,
    8.66,
    9.18,
    9.72,
    10.3,
    10.91,
    11.56,
    12.25,
    12.98,
    13.75,
    14.57,
    15.43,
    16.35,
    17.32,
    18.35,
    19.45,
    20.6,
    21.83,
    23.12,
    24.5,
    25.96,
    27.5,
    29.14,
    30.87,
    32.7,
    34.65,
    36.71,
    38.89,
    41.2,
    43.65,
    46.25,
    49,
    51.91,
    55,
    58.27,
    61.74,
    65.41,
    69.3,
    73.42,
    77.78,
    82.41,
    87.31,
    92.5,
    98,
    103.83,
    110,
    116.54,
    123.47,
    130.81,
    138.59,
    146.83,
    155.56,
    164.81,
    174.61,
    185,
    196,
    207.65,
    220,
    233.08,
    246.94,
    261.63,
    277.18,
    293.66,
    311.13,
    329.63,
    349.23,
    369.99,
    392,
    415.3,
    440,
    466.16,
    493.88,
    523.25,
    554.37,
    587.33,
    622.25,
    659.26,
    698.46,
    739.99,
    783.99,
    830.61,
    880,
    932.33,
    987.77,
    1046.5,
    1108.73,
    1174.66,
    1244.51,
    1318.51,
    1396.91,
    1479.98,
    1567.98,
    1661.22,
    1760,
    1864.66,
    1975.53,
    2093,
    2217.46,
    2349.32,
    2489.02,
    2637.02,
    2793.83,
    2959.96,
    3135.96,
    3322.44,
    3520,
    3729.31,
    3951.07,
    4186.01,
    4434.92,
    4698.64,
    4978.03,
    5274.04,
    5587.65,
    5919.91,
    6271.93,
    6644.88,
    7040,
    7458.62,
    7902.13,
    8372.02,
    8869.84,
    9397.27,
    9956.06,
    10548.08,
    11175.3,
    11839.82,
    12543.85
  ];
  var GeneralMIDIPercussion;
  (function(GeneralMIDIPercussion3) {
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["HighQ"] = 27] = "HighQ";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["Slap"] = 28] = "Slap";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["ScratchPush"] = 29] = "ScratchPush";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["ScratchPull"] = 30] = "ScratchPull";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["Sticks"] = 31] = "Sticks";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["SquareClick"] = 32] = "SquareClick";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["MetronomeClick"] = 33] = "MetronomeClick";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["MetronomeBell"] = 34] = "MetronomeBell";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["BassDrum2"] = 35] = "BassDrum2";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["BassDrum1"] = 36] = "BassDrum1";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["SideStick"] = 37] = "SideStick";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["SnareDrum1"] = 38] = "SnareDrum1";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["HandClap"] = 39] = "HandClap";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["SnareDrum2"] = 40] = "SnareDrum2";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["LowTom2"] = 41] = "LowTom2";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["ClosedHiHat"] = 42] = "ClosedHiHat";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["LowTom1"] = 43] = "LowTom1";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["PedalHiHat"] = 44] = "PedalHiHat";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["MidTom2"] = 45] = "MidTom2";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["OpenHiHat"] = 46] = "OpenHiHat";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["MidTom1"] = 47] = "MidTom1";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["HighTom2"] = 48] = "HighTom2";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["CrashCymbal1"] = 49] = "CrashCymbal1";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["HighTom1"] = 50] = "HighTom1";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["RideCymbal1"] = 51] = "RideCymbal1";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["ChineseCymbal"] = 52] = "ChineseCymbal";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["RideBell"] = 53] = "RideBell";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["Tambourine"] = 54] = "Tambourine";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["SplashCymbal"] = 55] = "SplashCymbal";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["Cowbell"] = 56] = "Cowbell";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["CrashCymbal2"] = 57] = "CrashCymbal2";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["VibraSlap"] = 58] = "VibraSlap";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["RideCymbal2"] = 59] = "RideCymbal2";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["HighBongo"] = 60] = "HighBongo";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["LowBongo"] = 61] = "LowBongo";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["MuteHighConga"] = 62] = "MuteHighConga";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["OpenHighConga"] = 63] = "OpenHighConga";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["LowConga"] = 64] = "LowConga";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["HighTimbale"] = 65] = "HighTimbale";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["LowTimbale"] = 66] = "LowTimbale";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["HighAgogo"] = 67] = "HighAgogo";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["LowAgogo"] = 68] = "LowAgogo";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["Cabasa"] = 69] = "Cabasa";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["Maracas"] = 70] = "Maracas";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["ShortWhistle"] = 71] = "ShortWhistle";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["LongWhistle"] = 72] = "LongWhistle";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["ShortGuiro"] = 73] = "ShortGuiro";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["LongGuiro"] = 74] = "LongGuiro";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["Claves"] = 75] = "Claves";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["HighWoodBlock"] = 76] = "HighWoodBlock";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["LowWoodBlock"] = 77] = "LowWoodBlock";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["MuteCuica"] = 78] = "MuteCuica";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["OpenCuica"] = 79] = "OpenCuica";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["MuteTriangle"] = 80] = "MuteTriangle";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["OpenTriangle"] = 81] = "OpenTriangle";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["Shaker"] = 82] = "Shaker";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["JingleBell"] = 83] = "JingleBell";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["Belltree"] = 84] = "Belltree";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["Castanets"] = 85] = "Castanets";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["MuteSurdo"] = 86] = "MuteSurdo";
    GeneralMIDIPercussion3[GeneralMIDIPercussion3["OpenSurdo"] = 87] = "OpenSurdo";
  })(GeneralMIDIPercussion || (GeneralMIDIPercussion = {}));
  var GeneralMidiInstruments;
  (function(GeneralMidiInstruments2) {
    GeneralMidiInstruments2[GeneralMidiInstruments2["AcousticGrand"] = 1] = "AcousticGrand";
    GeneralMidiInstruments2[GeneralMidiInstruments2["BrightAcoustic"] = 2] = "BrightAcoustic";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ElectricGrand"] = 3] = "ElectricGrand";
    GeneralMidiInstruments2[GeneralMidiInstruments2["HonkyTonk"] = 4] = "HonkyTonk";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ElectricPiano1"] = 5] = "ElectricPiano1";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ElectricPiano2"] = 6] = "ElectricPiano2";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Harpsichord"] = 7] = "Harpsichord";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Clav"] = 8] = "Clav";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Celesta"] = 9] = "Celesta";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Glockenspiel"] = 10] = "Glockenspiel";
    GeneralMidiInstruments2[GeneralMidiInstruments2["MusicBox"] = 11] = "MusicBox";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Vibraphone"] = 12] = "Vibraphone";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Marimba"] = 13] = "Marimba";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Xylophone"] = 14] = "Xylophone";
    GeneralMidiInstruments2[GeneralMidiInstruments2["TubularBells"] = 15] = "TubularBells";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Dulcimer"] = 16] = "Dulcimer";
    GeneralMidiInstruments2[GeneralMidiInstruments2["DrawbarOrgan"] = 17] = "DrawbarOrgan";
    GeneralMidiInstruments2[GeneralMidiInstruments2["PercussiveOrgan"] = 18] = "PercussiveOrgan";
    GeneralMidiInstruments2[GeneralMidiInstruments2["RockOrgan"] = 19] = "RockOrgan";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ChurchOrgan"] = 20] = "ChurchOrgan";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ReedOrgan"] = 21] = "ReedOrgan";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Accoridan"] = 22] = "Accoridan";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Harmonica"] = 23] = "Harmonica";
    GeneralMidiInstruments2[GeneralMidiInstruments2["TangoAccordian"] = 24] = "TangoAccordian";
    GeneralMidiInstruments2[GeneralMidiInstruments2["AcousticGuitar_nylon"] = 25] = "AcousticGuitar_nylon";
    GeneralMidiInstruments2[GeneralMidiInstruments2["AcousticGuitar_steel"] = 26] = "AcousticGuitar_steel";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ElectricGuitar_jazz"] = 27] = "ElectricGuitar_jazz";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ElectricGuitar_clean"] = 28] = "ElectricGuitar_clean";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ElectricGuitar_muted"] = 29] = "ElectricGuitar_muted";
    GeneralMidiInstruments2[GeneralMidiInstruments2["OverdrivenGuitar"] = 30] = "OverdrivenGuitar";
    GeneralMidiInstruments2[GeneralMidiInstruments2["DistortionGuitar"] = 31] = "DistortionGuitar";
    GeneralMidiInstruments2[GeneralMidiInstruments2["GuitarHarmonics"] = 32] = "GuitarHarmonics";
    GeneralMidiInstruments2[GeneralMidiInstruments2["AcousticBass"] = 33] = "AcousticBass";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ElectricBass_finger"] = 34] = "ElectricBass_finger";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ElectricBass_pick"] = 35] = "ElectricBass_pick";
    GeneralMidiInstruments2[GeneralMidiInstruments2["FretlessBass"] = 36] = "FretlessBass";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SlapBass1"] = 37] = "SlapBass1";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SlapBass2"] = 38] = "SlapBass2";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SynthBass1"] = 39] = "SynthBass1";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SynthBass2"] = 40] = "SynthBass2";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Violin"] = 41] = "Violin";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Viola"] = 42] = "Viola";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Cello"] = 43] = "Cello";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Contrabass"] = 44] = "Contrabass";
    GeneralMidiInstruments2[GeneralMidiInstruments2["TremoloStrings"] = 45] = "TremoloStrings";
    GeneralMidiInstruments2[GeneralMidiInstruments2["PizzicatoStrings"] = 46] = "PizzicatoStrings";
    GeneralMidiInstruments2[GeneralMidiInstruments2["OrchestralStrings"] = 47] = "OrchestralStrings";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Timpani"] = 48] = "Timpani";
    GeneralMidiInstruments2[GeneralMidiInstruments2["StringEnsemble1"] = 49] = "StringEnsemble1";
    GeneralMidiInstruments2[GeneralMidiInstruments2["StringEnsemble2"] = 50] = "StringEnsemble2";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SynthStrings1"] = 51] = "SynthStrings1";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SynthStrings2"] = 52] = "SynthStrings2";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ChoirAahs"] = 53] = "ChoirAahs";
    GeneralMidiInstruments2[GeneralMidiInstruments2["VoiceOohs"] = 54] = "VoiceOohs";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SynthVoice"] = 55] = "SynthVoice";
    GeneralMidiInstruments2[GeneralMidiInstruments2["OrchestraHit"] = 56] = "OrchestraHit";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Trumpet"] = 57] = "Trumpet";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Trombone"] = 58] = "Trombone";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Tuba"] = 59] = "Tuba";
    GeneralMidiInstruments2[GeneralMidiInstruments2["MutedTrumpet"] = 60] = "MutedTrumpet";
    GeneralMidiInstruments2[GeneralMidiInstruments2["FrenchHorn"] = 61] = "FrenchHorn";
    GeneralMidiInstruments2[GeneralMidiInstruments2["BrassSection"] = 62] = "BrassSection";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SynthBrass1"] = 63] = "SynthBrass1";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SynthBrass"] = 64] = "SynthBrass";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SopranoSax"] = 65] = "SopranoSax";
    GeneralMidiInstruments2[GeneralMidiInstruments2["AltoSax"] = 66] = "AltoSax";
    GeneralMidiInstruments2[GeneralMidiInstruments2["TenorSax"] = 67] = "TenorSax";
    GeneralMidiInstruments2[GeneralMidiInstruments2["BaritoneSax"] = 68] = "BaritoneSax";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Oboe"] = 69] = "Oboe";
    GeneralMidiInstruments2[GeneralMidiInstruments2["EnglishHorn"] = 70] = "EnglishHorn";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Bassoon"] = 71] = "Bassoon";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Clarinet"] = 72] = "Clarinet";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Piccolo"] = 73] = "Piccolo";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Flute"] = 74] = "Flute";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Recorder"] = 75] = "Recorder";
    GeneralMidiInstruments2[GeneralMidiInstruments2["PanFlute"] = 76] = "PanFlute";
    GeneralMidiInstruments2[GeneralMidiInstruments2["BlownBottle"] = 77] = "BlownBottle";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Shakuhachi"] = 78] = "Shakuhachi";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Whistle"] = 79] = "Whistle";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Ocarina"] = 80] = "Ocarina";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Lead1_square"] = 81] = "Lead1_square";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Lead2_sawtooth"] = 82] = "Lead2_sawtooth";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Lead3_calliope"] = 83] = "Lead3_calliope";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Lead4_chiff"] = 84] = "Lead4_chiff";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Lead5_charang"] = 85] = "Lead5_charang";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Lead6_voice"] = 86] = "Lead6_voice";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Lead7_fifths"] = 87] = "Lead7_fifths";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Lead8_bass_lead"] = 88] = "Lead8_bass_lead";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Pad1_new_age"] = 89] = "Pad1_new_age";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Pad2_warm"] = 90] = "Pad2_warm";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Pad3_polysynth"] = 91] = "Pad3_polysynth";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Pad4_choir"] = 92] = "Pad4_choir";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Pad5_bowed"] = 93] = "Pad5_bowed";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Pad6_metallic"] = 94] = "Pad6_metallic";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Pad7_halo"] = 95] = "Pad7_halo";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Pad8_sweep"] = 96] = "Pad8_sweep";
    GeneralMidiInstruments2[GeneralMidiInstruments2["FX1_rain"] = 97] = "FX1_rain";
    GeneralMidiInstruments2[GeneralMidiInstruments2["FX2_soundtrack"] = 98] = "FX2_soundtrack";
    GeneralMidiInstruments2[GeneralMidiInstruments2["FX3_crystal"] = 99] = "FX3_crystal";
    GeneralMidiInstruments2[GeneralMidiInstruments2["FX4_atmosphere"] = 100] = "FX4_atmosphere";
    GeneralMidiInstruments2[GeneralMidiInstruments2["FX5_brightness"] = 101] = "FX5_brightness";
    GeneralMidiInstruments2[GeneralMidiInstruments2["FX6_goblins"] = 102] = "FX6_goblins";
    GeneralMidiInstruments2[GeneralMidiInstruments2["FX7_echoes"] = 103] = "FX7_echoes";
    GeneralMidiInstruments2[GeneralMidiInstruments2["FX8_sci_fi"] = 104] = "FX8_sci_fi";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Sitar"] = 105] = "Sitar";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Banjo"] = 106] = "Banjo";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Shamisen"] = 107] = "Shamisen";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Koto"] = 108] = "Koto";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Kalimba"] = 109] = "Kalimba";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Bagpipe"] = 110] = "Bagpipe";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Fiddle"] = 111] = "Fiddle";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Shanai"] = 112] = "Shanai";
    GeneralMidiInstruments2[GeneralMidiInstruments2["TinkleBell"] = 113] = "TinkleBell";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Agogo"] = 114] = "Agogo";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SteelDrums"] = 115] = "SteelDrums";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Woodblock"] = 116] = "Woodblock";
    GeneralMidiInstruments2[GeneralMidiInstruments2["TaikoDrum"] = 117] = "TaikoDrum";
    GeneralMidiInstruments2[GeneralMidiInstruments2["MelodicTom"] = 118] = "MelodicTom";
    GeneralMidiInstruments2[GeneralMidiInstruments2["SynthDrum"] = 119] = "SynthDrum";
    GeneralMidiInstruments2[GeneralMidiInstruments2["ReverseCymbal"] = 120] = "ReverseCymbal";
    GeneralMidiInstruments2[GeneralMidiInstruments2["GuitarFretNoise"] = 121] = "GuitarFretNoise";
    GeneralMidiInstruments2[GeneralMidiInstruments2["BreathNoise"] = 122] = "BreathNoise";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Seashore"] = 123] = "Seashore";
    GeneralMidiInstruments2[GeneralMidiInstruments2["BirdTweet"] = 124] = "BirdTweet";
    GeneralMidiInstruments2[GeneralMidiInstruments2["TelephoneRing"] = 125] = "TelephoneRing";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Helicopter"] = 126] = "Helicopter";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Applause"] = 127] = "Applause";
    GeneralMidiInstruments2[GeneralMidiInstruments2["Gunshot"] = 128] = "Gunshot";
  })(GeneralMidiInstruments || (GeneralMidiInstruments = {}));

  // src/midi/audio.ts
  var DURATION_CLICK_PAD = 0.015;
  function createCompressor(system2, vpc, _length, time) {
    const compressor = system2.context().createDynamicsCompressor();
    compressor.threshold.setValueAtTime(vpc.threshold, time);
    compressor.knee.setValueAtTime(vpc.knee, time);
    compressor.ratio.setValueAtTime(vpc.ratio, time);
    compressor.attack.setValueAtTime(vpc.attack, time);
    compressor.release.setValueAtTime(vpc.release, time);
    return compressor;
  }
  function createADSR(system2, padsr, length, time) {
    const asdr = system2.context().createGain();
    const now = time;
    const atkDuration = padsr.attack * length;
    const atkEndTime = now + atkDuration;
    const decayDuration = padsr.decay * length;
    asdr.gain.setValueAtTime(1e-5, now);
    asdr.gain.exponentialRampToValueAtTime(1, atkEndTime);
    asdr.gain.exponentialRampToValueAtTime(padsr.sustain, now + decayDuration);
    const relDuration = padsr.release * length;
    const relEndTime = now + relDuration;
    asdr.gain.exponentialRampToValueAtTime(padsr.sustain, now + relDuration);
    asdr.gain.exponentialRampToValueAtTime(1e-5, now + length + DURATION_CLICK_PAD);
    return asdr;
  }
  function createFilter(system2, pf, _length, time) {
    const filter = system2.context().createBiquadFilter();
    filter.type = pf.type;
    filter.frequency.setValueAtTime(pf.frequency, time);
    filter.detune.setValueAtTime(pf?.detune ?? 1, time);
    filter.Q.setValueAtTime(pf.q, time);
    filter.gain.setValueAtTime(pf.gain, time);
    return filter;
  }
  function patchPlayer(system2, patch) {
    return (note, time, output) => {
      const midi = NoteToMidi.get(note);
      if (!midi) {
        console.warn(`Could not find midi value for note: ${note} Bad note?`);
        return NoOp;
      }
      const fundamental = MidiToFreq[midi];
      if (!fundamental)
        throw Error(`Could not find frequency value for note. Bad midi? ${midi}`);
      let oscillatorOutput;
      const soundLen = patch.len;
      const patchGain = system2.context().createGain();
      const sampleGain = patch.vca?.gain ?? 1;
      patchGain.gain.linearRampToValueAtTime(sampleGain, time);
      const processing = new Array();
      if (patch.filter) {
        processing.push(createFilter(system2, patch.filter, soundLen, time));
      }
      if (patch.adsr) {
        processing.push(createADSR(system2, patch.adsr, soundLen, time));
      }
      if (patch.compressor) {
        processing.push(createCompressor(system2, patch.compressor, soundLen, time));
      }
      processing.push(patchGain);
      if (processing.length > 1) {
        for (let z = 1; z < processing.length; z++) {
          const fx = processing[z];
          processing[z - 1].connect(fx);
        }
      }
      oscillatorOutput = processing[0];
      const vcoCleanUp = [];
      const voiceLen = Math.min(patch.voices.length, 8);
      for (let v = 0; v < voiceLen; v++) {
        const voice = patch.voices[v];
        let vco;
        if (voice.vco.type === "custom") {
          vco = new AudioWorkletNode(system2.context(), "white-noise-processor");
        } else {
          vco = system2.context().createOscillator();
          const lvco = vco;
          const voice2 = patch.voices[v];
          lvco.type = voice2.vco.type;
          if (voice2.vco.frequency) {
            lvco.frequency.value = voice2.vco.frequency;
          } else if (voice2.vco.freqOffset) {
            lvco.frequency.value = fundamental + voice2.vco.freqOffset;
          } else if (voice2.vco.freqMult) {
            lvco.frequency.value = fundamental * voice2.vco.freqMult;
          } else {
            lvco.frequency.value = fundamental;
          }
          lvco.detune.value = voice2.vco.detune;
          lvco.start(time);
          if (soundLen > 0)
            lvco.stop(time + soundLen + DURATION_CLICK_PAD * 10);
        }
        const vca = system2.context().createGain();
        const pvca = patch.voices[v].vca;
        let vcaVolume = 1;
        if (pvca.gainOffset) {
          vcaVolume = 1 + pvca.gainOffset;
        }
        if (pvca.gain) {
          vcaVolume = pvca.gain;
        }
        vca.gain.setValueAtTime(vcaVolume, time);
        vco.connect(vca);
        vca.connect(oscillatorOutput);
        if (soundLen > 0) {
          vca.gain.exponentialRampToValueAtTime(vcaVolume, time + soundLen - DURATION_CLICK_PAD);
          vca.gain.exponentialRampToValueAtTime(1e-5, time + soundLen + DURATION_CLICK_PAD);
        }
        vcoCleanUp.push((t) => {
          vca.gain.exponentialRampToValueAtTime(1e-5, t);
          if (vco?.stop) {
            vco?.stop();
          }
          vco.disconnect(vca);
          vca.disconnect(oscillatorOutput);
        });
      }
      patchGain.connect(output);
      return (t) => {
        const time2 = t ?? system2.currentTime();
        patchGain.gain.setValueAtTime(patchGain.gain.value, time2);
        patchGain.gain.linearRampToValueAtTime(0, time2 + DURATION_CLICK_PAD * 22);
        setTimeout(() => {
          vcoCleanUp.forEach((cleanUp) => cleanUp(time2 + DURATION_CLICK_PAD * 10));
          if (processing.length > 1) {
            for (let z = 1; z < processing.length; z++) {
              const fx = processing[z];
              processing[z - 1].disconnect(fx);
            }
          }
          patchGain.disconnect(output);
        }, 333);
      };
    };
  }

  // src/midi/andy-patch.json
  var andy_patch_default = [
    [1, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [2, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [3, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [4, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [5, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [6, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [7, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [8, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [9, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [10, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [11, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [12, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [13, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [14, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [15, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [16, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [17, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [18, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [19, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [20, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [21, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [22, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [23, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [24, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [25, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [26, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [27, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [28, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [29, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [30, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [31, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [32, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [33, {len: 3, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "triangle", detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "triangle", detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "square", detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -10, knee: 0, ratio: 1, attack: 0, release: 0.5}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 300, q: 1.04, gain: 1}}],
    [34, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [35, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [36, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [37, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [38, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [39, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [40, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [41, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [42, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [43, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [44, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [45, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [46, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [47, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [48, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [49, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [50, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [51, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [52, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [53, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [54, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [55, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [56, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [57, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [58, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [59, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [60, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [61, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [62, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [63, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [64, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [65, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [66, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [67, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [68, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [69, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [70, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [71, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [72, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [73, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [74, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [75, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [76, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [77, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [78, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [79, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [80, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [81, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [82, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [83, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [84, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [85, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [86, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [87, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [88, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [89, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [90, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [91, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [92, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [93, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [94, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [95, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [96, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [97, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [98, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [99, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [100, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [101, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [102, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [103, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [104, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [105, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [106, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [107, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [108, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [109, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [110, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [111, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [112, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [113, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [114, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [115, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [116, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [117, {len: 2, vca: {gain: 5}, voices: [{vco: {type: "custom", freqOffset: 0, detune: 0}, vca: {gainOffset: 1.5}}, {vco: {type: "square", freqMult: 1.1, detune: 0}, vca: {gainOffset: -0.79}}], compressor: {threshold: -1, knee: 40, ratio: 2, attack: 0.2, release: 0.3}, adsr: {attack: 0, decay: 0.1, sustain: 0.4, release: 0.1}, filter: {type: "lowpass", frequency: 100, q: 1.04, gain: 2}}],
    [118, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [119, {len: 2, vca: {gain: 2}, voices: [{vco: {type: "custom", freqOffset: 0, detune: 0}, vca: {gainOffset: 2}}], compressor: {threshold: -10, knee: 0, ratio: 2, attack: 0, release: 0}, adsr: {attack: 0, decay: 0, sustain: 1, release: 1}, filter: {type: "lowpass", frequency: 80, q: 1.04, gain: 0.5}}],
    [120, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [121, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [122, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [123, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [124, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [125, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [126, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [127, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [128, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}]
  ];

  // src/midi/andy-percus.json
  var andy_percus_default = [
    [27, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [28, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [29, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [30, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [31, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [32, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [33, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [34, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [35, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [36, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [37, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [38, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [39, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [30, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [31, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [32, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [33, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [34, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [35, {len: 0.2, vca: {gain: 2.5}, voices: [{vco: {type: "custom", freqOffset: 0, detune: 0}, vca: {gainOffset: 10}}, {vco: {type: "triangle", frequency: 80, detune: 0}, vca: {gainOffset: 10}}], compressor: {threshold: -30, knee: 40, ratio: 12, attack: 0, release: 0.25}, adsr: {attack: 0, decay: 0.01, sustain: 0.5, release: 0.01}, filter: {type: "lowpass", frequency: 200, q: 1.04, gain: 10}}],
    [36, {len: 0.5, vca: {gain: 3}, voices: [{vco: {type: "custom", freqOffset: 0, detune: 0}, vca: {gainOffset: 1.5}}, {vco: {type: "square", frequency: 88, detune: 0}, vca: {gainOffset: -0.79}}], compressor: {threshold: -1, knee: 40, ratio: 2, attack: 0.2, release: 0.3}, adsr: {attack: 0, decay: 0.1, sustain: 0.4, release: 0.1}, filter: {type: "lowpass", frequency: 80, q: 1.04, gain: 2}}],
    [37, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [38, {len: 0.2, vca: {gain: 8}, voices: [{vco: {type: "custom", freqOffset: 0, detune: 0}, vca: {gainOffset: 10}}, {vco: {type: "triangle", frequency: 150, detune: -10}, vca: {gainOffset: 2}}], compressor: {threshold: -30, knee: 40, ratio: 12, attack: 0, release: 0.25}, adsr: {attack: 0, decay: 0.01, sustain: 0.5, release: 0.01}, filter: {type: "lowpass", frequency: 1e3, q: 1.04, gain: 10}}],
    [39, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [40, {len: 0.5, vca: {gain: 3.5}, voices: [{vco: {type: "custom", freqOffset: 0, detune: 0}, vca: {gainOffset: 1.5}}], compressor: {threshold: -20, knee: 30, ratio: 2, attack: 0.2, release: 0.3}, adsr: {attack: 0, decay: 0.1, sustain: 0.4, release: 0.1}, filter: {type: "bandpass", frequency: 180, q: 3.04, gain: 0.3}}],
    [41, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [42, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [43, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [44, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [45, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [46, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [47, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [48, {len: 0.5, vca: {gain: 3}, voices: [{vco: {type: "custom", freqOffset: 0, detune: 0}, vca: {gainOffset: 1.5}}, {vco: {type: "square", freqMult: 1.1, detune: 0}, vca: {gainOffset: -0.79}}], compressor: {threshold: -1, knee: 40, ratio: 2, attack: 0.2, release: 0.3}, adsr: {attack: 0, decay: 0.1, sustain: 0.4, release: 0.1}, filter: {type: "lowpass", frequency: 80, q: 1.04, gain: 2}}],
    [49, {len: 1.5, vca: {gain: 2.5}, voices: [{vco: {type: "triangle", frequency: 4740, detune: 0}, vca: {gainOffset: -0.9}}, {vco: {type: "triangle", frequency: 692, detune: 0}, vca: {gainOffset: -0.9}}, {vco: {type: "triangle", frequency: 1783, detune: 0}, vca: {gainOffset: -0.9}}, {vco: {type: "custom", frequency: 0, detune: 0}, vca: {gainOffset: 0.9}}], filter: {type: "highpass", frequency: 3349, q: 1.04, gain: 1}, adsr: {attack: 0, decay: 0.2, sustain: 0.05, release: 0.2}, compressor: {threshold: -20, knee: 40, ratio: 12, attack: 0.1, release: 0.2}}],
    [50, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [51, {len: 1.5, vca: {gain: 2.5}, voices: [{vco: {type: "sine", frequency: 174, detune: 0}, vca: {gainOffset: -0.9}}, {vco: {type: "sine", frequency: 692, detune: 0}, vca: {gainOffset: -0.9}}, {vco: {type: "sine", frequency: 1783, detune: 0}, vca: {gainOffset: -0.9}}, {vco: {type: "custom", frequency: 220, detune: 0}, vca: {gainOffset: 0.9}}], filter: {type: "bandpass", frequency: 1349, q: 1.04, gain: 1}, adsr: {attack: 0, decay: 0.2, sustain: 0.05, release: 0.2}, compressor: {threshold: -20, knee: 40, ratio: 12, attack: 0.1, release: 0.2}}],
    [52, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [53, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [54, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [55, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [56, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [57, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [58, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [59, {len: 1, vca: {gain: 3}, voices: [{vco: {type: "triangle", frequency: 1740, detune: 0}, vca: {gainOffset: -0.8}}, {vco: {type: "triangle", frequency: 692, detune: 0}, vca: {gainOffset: -0.7}}, {vco: {type: "triangle", frequency: 1783, detune: 0}, vca: {gainOffset: -0.9}}, {vco: {type: "custom", frequency: 0, detune: 0}, vca: {gainOffset: 0.3}}], filter: {type: "bandpass", frequency: 8349, q: 1.04, gain: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.05, release: 0.1}, compressor: {threshold: -20, knee: 40, ratio: 12, attack: 0.1, release: 0.2}}],
    [60, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [61, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [62, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [63, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [64, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [65, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [66, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [67, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [68, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [69, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [70, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [71, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [72, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [73, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [74, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [75, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [76, {len: 1, vca: {gain: 3}, voices: [{vco: {type: "custom", freqOffset: 0, detune: 0}, vca: {gainOffset: 0.9}}, {vco: {type: "triangle", frequency: 1e3, detune: 0}, vca: {gainOffset: -0.9}}], filter: {type: "bandpass", frequency: 1240, q: 10.94, gain: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.1, release: 0.2}, compressor: {threshold: -20, knee: 30, ratio: 2, attack: 0.1, release: 0.4}}],
    [77, {len: 0.5, vca: {gain: 3}, voices: [{vco: {type: "custom", freqOffset: 0, detune: 0}, vca: {gainOffset: 1.5}}, {vco: {type: "square", freqMult: 1.1, detune: 0}, vca: {gainOffset: -0.79}}], compressor: {threshold: -1, knee: 40, ratio: 2, attack: 0.2, release: 0.3}, adsr: {attack: 0, decay: 0.1, sustain: 0.4, release: 0.1}, filter: {type: "bandpass", frequency: 640, q: 10.04, gain: 0.3}}],
    [78, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [79, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [80, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [81, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [82, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [83, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [84, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [85, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [86, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}],
    [87, {len: 20, vca: {gain: 0.2}, voices: [{vco: {type: "sine", freqOffset: 0, detune: 0}, vca: {gainOffset: 0}}, {vco: {type: "sine", freqMult: 2, detune: 0}, vca: {gainOffset: -0.29}}, {vco: {type: "sine", freqMult: 4, detune: 0}, vca: {gainOffset: -0.4}}, {vco: {type: "sine", freqMult: 8, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 16, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 32, detune: 0}, vca: {gainOffset: -0.62}}, {vco: {type: "sine", freqMult: 64, detune: 0}, vca: {gainOffset: -0.68}}, {vco: {type: "sine", freqMult: 128, detune: 0}, vca: {gainOffset: -0.56}}], compressor: {threshold: -20, knee: 0, ratio: 12, attack: 0, release: 1}, adsr: {attack: 0, decay: 0.1, sustain: 0.5, release: 1}, filter: {type: "lowpass", frequency: 2300, q: 1.04, gain: 1}}]
  ];

  // src/midi/patchBank.ts
  var AndyPatch = new Map(andy_patch_default);
  var AndyPercussionPatch = new Map(andy_percus_default);

  // src/io/file.ts
  var BinaryFile = class {
    constructor(data) {
      this.buffer = data;
      this.dataView = new DataView(this.buffer, 0);
      this.offset = 0;
    }
    tell() {
      return this.offset;
    }
    seek(offset) {
      this.offset = offset;
    }
    skip(bytes) {
      this.offset += bytes;
    }
    eof() {
      return this.offset >= this.buffer.byteLength;
    }
    slice(offset, length) {
      return new BinaryFile(this.buffer.slice(offset, offset + length));
    }
    read(length) {
      const result = new BinaryFile(this.buffer.slice(this.offset, this.offset + length));
      this.offset += length;
      return result;
    }
    readString(length) {
      let result = "";
      let terminated = false;
      for (let i = 0; i < length; i++) {
        const byte = this.readUInt8();
        if (byte === 0) {
          terminated = true;
        }
        if (!terminated) {
          result += String.fromCharCode(byte);
        }
      }
      return result;
    }
    readCString() {
      let result = "";
      for (let i = 0; i < 128; i++) {
        const byte = this.readUInt8();
        if (byte === 0) {
          return result;
        }
        result += String.fromCharCode(byte);
      }
      return result;
    }
    readChars(count) {
      return this.readBytes(count);
    }
    readBytes(count) {
      let result = new Uint8Array(count);
      for (let i = 0; i < count; i++) {
        const byte = this.readUInt8();
        result[i] = byte;
      }
      return result;
    }
    readSBytes(count) {
      let result = new Int8Array(count);
      for (let i = 0; i < count; i++) {
        const byte = this.readInt8();
        result[i] = byte;
      }
      return result;
    }
    readUInt8() {
      return this.dataView.getUint8(this.offset++);
    }
    readInt8() {
      return this.dataView.getInt8(this.offset++);
    }
    readUInt16(little = true) {
      const result = this.dataView.getUint16(this.offset, little);
      this.offset += 2;
      return result;
    }
    readInt16(little = true) {
      const result = this.dataView.getInt16(this.offset, little);
      this.offset += 2;
      return result;
    }
    readUInt32(little = true) {
      const result = this.dataView.getUint32(this.offset, little);
      this.offset += 4;
      return result;
    }
    readInt32(little = true) {
      const result = this.dataView.getInt32(this.offset, little);
      this.offset += 4;
      return result;
    }
    readFloat(little = true) {
      const result = this.dataView.getFloat32(this.offset, little);
      this.offset += 4;
      return result;
    }
    close() {
      this.offset = -1;
    }
  };

  // src/midi/song.ts
  var ChunkType;
  (function(ChunkType2) {
    ChunkType2["Header"] = "MThd";
    ChunkType2["Track"] = "MTrk";
  })(ChunkType || (ChunkType = {}));
  var Format;
  (function(Format2) {
    Format2[Format2["Single"] = 0] = "Single";
    Format2[Format2["MultiSync"] = 1] = "MultiSync";
    Format2[Format2["MultiAsync"] = 2] = "MultiAsync";
  })(Format || (Format = {}));
  var MIDI = class {
    constructor(data) {
      this.markers = new Map([]);
      this.data = new BinaryFile(data);
      this.format = 0;
      this.numberOfTracks = 0;
      this.ticksPerQuarterNote = 0;
      this.tracks = new Array(this.numberOfTracks);
      this.readHeader();
      for (let t = 0; t < this.numberOfTracks; t++) {
        this.tracks[t] = this.readTrack();
      }
    }
    readHeader() {
      const hdr = this.data.readString(4);
      const size = this.data.readUInt32(false);
      if (hdr !== ChunkType.Header || size !== 6) {
        throw new Error(`Incorrect header, not a midi file? ${hdr} ${size}`);
      }
      const format = this.data.readUInt16(false);
      this.format = format;
      const tracks = this.data.readUInt16(false);
      this.numberOfTracks = tracks;
      const ticks = this.data.readUInt16(false);
      if ((ticks & 32768) === 0) {
        this.ticksPerQuarterNote = ticks;
      } else {
        throw new Error("Divisions set as units per SMTPE frame: Unsupported");
      }
    }
    readTrack() {
      const events = [];
      const track = {
        name: "Unknown",
        events
      };
      const hdr = this.data.readString(4);
      const length = this.data.readUInt32(false);
      const start = this.data.tell();
      let lastEvent = 0;
      let trackTime = 0;
      while (true) {
        let dtime = this.readVariableLen();
        let eventType = this.data.readUInt8();
        trackTime += dtime;
        if (this.data.tell() >= start + length) {
          break;
        }
        if ((eventType & 240) === 240) {
          if (eventType === 255) {
            const event = this.metaEvents(dtime, trackTime, eventType, track);
            events.push(event);
            if (event.raw === 47) {
              break;
            }
          } else if (eventType === 240 || eventType === 247) {
            const event = this.sysexEvents(dtime, trackTime, eventType);
            events.push(event);
          } else {
            console.warn(`Unknown event ${eventType.toString(16)}`);
          }
        } else {
          const event = this.midiEvents(dtime, trackTime, eventType, lastEvent);
          lastEvent = event.raw;
          events.push(event);
        }
      }
      return track;
    }
    metaEvents(dtime, ttime, eventType, track) {
      const event = {
        raw: eventType,
        time: dtime,
        abstime: ttime,
        type: "Meta",
        data: []
      };
      const command = this.data.readUInt8();
      const length = this.readVariableLen();
      event.raw = command;
      switch (command) {
        case 0:
          {
            event.subtype = "SequenceNumber";
            event.data?.push(this.data.readUInt16(false));
          }
          break;
        case 1:
          {
            event.subtype = "Text";
            const txt = this.data.readString(length);
            event.data?.push(txt);
          }
          break;
        case 2:
          {
            event.subtype = "CopyrightNotice";
            const copy = this.data.readString(length);
            event.data?.push(copy);
          }
          break;
        case 3:
          {
            event.subtype = "TrackName";
            const name = this.data.readString(length);
            event.data?.push(name);
            track.name = name;
          }
          break;
        case 4:
          {
            event.subtype = "InstrumentName";
            const name = this.data.readString(length);
            event.data?.push(name);
          }
          break;
        case 5:
          {
            event.subtype = "Lyrics";
            const lyrics = this.data.readString(length);
            event.data?.push(lyrics);
          }
          break;
        case 6:
          {
            event.subtype = "Marker";
            const markerStr = this.data.readString(length);
            event.data?.push(markerStr);
            this.markers.set(markerStr, {time: dtime, abstime: ttime});
          }
          break;
        case 7:
          {
            event.subtype = "CuePoint";
            const cueStr = this.data.readString(length);
            event.data?.push(cueStr);
          }
          break;
        case 15:
          {
            event.subtype = "Unknown";
            if (length > 0) {
              const str = this.data.readBytes(length);
              event.data?.push(str);
            }
          }
          break;
        case 32:
          {
            event.subtype = "MidiChannelPrefix";
            event.data?.push(this.data.readUInt8());
          }
          break;
        case 33:
          {
            event.subtype = "Unknown";
            event.data?.push(this.data.readUInt8());
          }
          break;
        case 47:
          {
            event.subtype = "EndOfTrack";
          }
          break;
        case 81:
          {
            event.subtype = "SetTempo";
            const microseconds = (this.data.readUInt8() << 16) + (this.data.readUInt8() << 8) + this.data.readUInt8();
            event.data?.push(microseconds);
          }
          break;
        case 84:
          {
            event.subtype = "SMPTEOffset";
            if (length > 0)
              this.data.readBytes(length);
          }
          break;
        case 88:
          {
            event.subtype = "TimeSignature";
            event.data?.push(this.data.readUInt8());
            event.data?.push(Math.pow(2, this.data.readUInt8()));
            event.data?.push(this.data.readUInt8());
            event.data?.push(this.data.readUInt8());
          }
          break;
        case 89:
          {
            event.subtype = "KeySignature";
            event.data?.push(this.data.readInt8());
            event.data?.push(this.data.readUInt8());
          }
          break;
        case 127:
          {
            event.subtype = "SequencerSpecific";
            const _data = this.data.readBytes(length);
          }
          break;
        default: {
          console.warn(`Unknown meta event ${eventType.toString(16)} ${command.toString(16)} ${length}`);
          event.subtype = "Unknown";
          if (length > 0)
            event.data?.push(this.data.readBytes(length));
        }
      }
      return event;
    }
    sysexEvents(dtime, ttime, eventType) {
      const event = {
        raw: eventType,
        abstime: ttime,
        time: dtime,
        type: "SysEx",
        data: []
      };
      const length = this.readVariableLen();
      const _data = this.data.readBytes(length);
      return event;
    }
    midiEvents(dtime, ttime, command, lastEvent) {
      let useEvent = 0;
      let param = 0;
      if ((command & 128) === 0) {
        param = command;
        useEvent = lastEvent;
      } else {
        param = this.data.readUInt8();
        useEvent = command;
      }
      const event = {
        raw: useEvent,
        time: dtime,
        abstime: ttime,
        type: "Channel",
        data: []
      };
      const cmd = (useEvent & 240) >> 4;
      const channel = useEvent & 15;
      event.channel = channel;
      switch (cmd) {
        case 8:
          {
            event.subtype = "NoteOff";
            const note = param;
            event.data?.push(note);
            const velocity = this.data.readUInt8();
            event.data?.push(velocity);
          }
          break;
        case 9:
          {
            const note = param;
            event.data?.push(note);
            const velocity = this.data.readUInt8();
            event.data?.push(velocity);
            if (velocity == 0)
              event.subtype = "NoteOff";
            else
              event.subtype = "NoteOn";
          }
          break;
        case 10:
          {
            event.subtype = "NoteAfterTouch";
            event.data?.push(param);
            const amount = this.data.readUInt8();
            event.data?.push(amount);
          }
          break;
        case 11:
          {
            event.subtype = "Controller";
            event.data?.push(param);
            const value = this.data.readUInt8();
            event.data?.push(value);
          }
          break;
        case 12:
          {
            event.subtype = "ProgramChange";
            const type = param;
            event.data?.push(type);
          }
          break;
        case 13:
          {
            event.subtype = "ChannelAfterTouch";
            event.data?.push(param);
          }
          break;
        case 14:
          {
            event.subtype = "PitchBend";
            const amount = param + (this.data.readUInt8() << 7);
            event.data?.push(amount);
          }
          break;
        default:
          console.warn(`[[UNKNOWN ${command.toString(16)}]]`);
      }
      return event;
    }
    readVariableLen() {
      const start = this.data.readUInt8();
      if ((start & 128) === 0) {
        return start;
      }
      let c = 0;
      let value = start & 127;
      do {
        value = (value << 7) + (c = this.data.readUInt8() & 127);
      } while (c & 0);
      return value;
    }
  };
  var loadMidi = (url) => {
    return new Promise((resolve, reject) => {
      const headers = new Headers();
      headers.append("User-Agent", "Andy/1.0");
      headers.append("Accept", "audio/midi");
      const body = void 0;
      fetch(url, {
        method: "GET",
        mode: "cors",
        cache: "reload",
        headers,
        credentials: "same-origin",
        redirect: "follow",
        referrerPolicy: "no-referrer",
        body: body ? JSON.stringify(body) : body
      }).then((response) => response.arrayBuffer()).then((response) => {
        const mid = new MIDI(response);
        resolve(mid);
      }).catch((e) => console.error(e));
    });
  };

  // src/metronome.ts
  var system;
  var isPlaying = false;
  var lookahead = 25;
  var scheduleAheadTime = 0.1;
  var nextNoteTime = 0;
  var timerWorker;
  var beatInfo;
  var totalTick = 0;
  function nextTick(tempo, sig) {
    const secondsPerQuaterNote = 6e4 / (tempo.bpm * tempo.ppq);
    const millisecondsPerQuarterNote = secondsPerQuaterNote / 1e3;
    nextNoteTime += millisecondsPerQuarterNote;
  }
  var toSignature = (sig) => {
    const p = sig.split("/");
    return createSignature(parseInt(p[0]), parseInt(p[1]));
  };
  var createSignature = (numerator, denominator, quarter32nd, metroTicks = 24) => {
    return {
      noteBeats: numerator,
      perMeasure: denominator,
      ticksInMetronome: metroTicks,
      toQuarter32nd: quarter32nd ?? numerator * 2
    };
  };
  var createTempo = (ppq, usPerQuarter) => {
    const tempo = {
      ppq,
      tempo: usPerQuarter,
      bpm: 6e7 / (usPerQuarter > 0 ? usPerQuarter : 0)
    };
    return tempo;
  };
  function scheduler(tempo, sig, perTick) {
    while (nextNoteTime < system.currentTime() + scheduleAheadTime) {
      if (totalTick % tempo.ppq === 0) {
        beatInfo.beat++;
        beatInfo.tick = 0;
      } else {
        beatInfo.tick += 1;
      }
      if (beatInfo.beat === sig.noteBeats + 1) {
        beatInfo.measure++;
        beatInfo.beat = 1;
        beatInfo.tick = 0;
      }
      for (let l = 0; l < perTick.tickListeners.length; l++) {
        [tempo, sig] = perTick.tickListeners[l](totalTick, nextNoteTime, beatInfo);
      }
      nextTick(tempo, sig);
      totalTick++;
    }
  }
  function playHeadToggle() {
    isPlaying = !isPlaying;
    if (isPlaying) {
      nextNoteTime = system.currentTime();
      timerWorker.postMessage("start");
    } else {
      timerWorker.postMessage("stop");
    }
  }
  function playHeadTo(time) {
    totalTick = time;
  }
  function playHeadRewind() {
    nextNoteTime = system.currentTime();
    beatInfo = {
      measure: 1,
      beat: 1,
      tick: 0
    };
    totalTick = 0;
  }
  function createPlayHeadClock(sys, tempo, sig, perTick) {
    system = sys;
    timerWorker = new Worker("metronome.worker.js");
    playHeadRewind();
    timerWorker.onmessage = function(e) {
      if (e.data === "tick") {
        scheduler(tempo, sig, perTick);
      } else {
        console.log("message: " + e.data);
      }
    };
    timerWorker.postMessage({"interval": lookahead});
  }

  // src/midi/player.ts
  var MAX_CHANNELS = 16;
  function createRun(system2, song) {
    if (song.format === Format.MultiAsync)
      throw new Error("Currently only support synchronous tracks");
    const run = {
      song,
      tracks: [],
      output: system2.context().createGain(),
      tickListeners: []
    };
    for (let i = 0; i < song.numberOfTracks; i++) {
      const pan = system2.context().createPanner();
      pan.orientationX.setValueAtTime(0, 0);
      pan.orientationY.setValueAtTime(0, 0);
      pan.orientationZ.setValueAtTime(-1, 0);
      pan.positionX.setValueAtTime(0, 0);
      pan.positionY.setValueAtTime(1, 0);
      pan.positionZ.setValueAtTime(0, 0);
      const vol = system2.context().createGain();
      pan.connect(vol);
      vol.connect(run.output);
      vol.gain.setValueAtTime(1, 0);
      const startFns = new Array(MAX_CHANNELS);
      for (let i2 = 0; i2 < MAX_CHANNELS; i2++) {
        startFns[i2] = [];
      }
      const t = {
        lastFn: startFns,
        idx: 0,
        vol,
        pan
      };
      run.tracks.push(t);
    }
    run.output.connect(system2.masterGainNode());
    return run;
  }
  function destroyRun(system2, run) {
    run.output.disconnect(system2.masterGainNode());
    while (run.tickListeners.length) {
      run.tickListeners.pop();
    }
  }
  var useInstrument = (system2, run, patchNum) => {
    const patch = AndyPatch.get(patchNum);
    if (patch)
      return patchPlayer(system2, patch);
    else {
      const p = AndyPatch.get(GeneralMidiInstruments.ChurchOrgan);
      if (p)
        return patchPlayer(system2, p);
    }
    return NoOp;
  };
  var usePercussion = (system2, run, patchNum) => {
    const patch = AndyPercussionPatch.get(patchNum);
    if (patch)
      return patchPlayer(system2, patch);
    return NoOp;
  };
  function silenceRun(songRun) {
    songRun.tracks.forEach((track) => {
      for (let i = 0; i < MAX_CHANNELS; i++) {
        const fns = track.lastFn[i];
        while (fns.length) {
          const soundOff = fns.pop();
          if (soundOff)
            soundOff();
        }
      }
    });
  }
  function setRunTimeTo(time, songRun) {
    const len = songRun.song.tracks.length;
    for (let t = 0; t < len; t++) {
      const elen = songRun.song.tracks[t].events.length;
      for (let e = 0; e < elen; e++) {
        const abstime = songRun.song.tracks[t].events[e].abstime;
        if (abstime >= time) {
          songRun.tracks[t].idx = e;
          break;
        }
      }
    }
  }
  function rewindRun(songRun) {
    songRun.tracks.forEach((track) => {
      track.idx = 0;
    });
  }
  function handleEventCallback(system2, run) {
    let tempo = createTempo(run.song.ticksPerQuarterNote, 5e5);
    let sig = toSignature("4/4");
    const tickCallBack = (tick, time, bi) => {
      const trackLen = run.song.tracks.length;
      const stopAllSounds = (track) => {
        for (let i = 0; i < MAX_CHANNELS; i++) {
          const fns = track.lastFn[i];
          while (fns.length) {
            const soundOff = fns.pop();
            if (soundOff)
              soundOff();
          }
        }
      };
      for (let t = 0; t < trackLen; t++) {
        const track = run.tracks[t];
        const events = run.song.tracks[t].events;
        for (let e = track.idx; e < events.length; e++) {
          const ce = events[e];
          if (ce.abstime !== tick) {
            track.idx = e;
            break;
          }
          if (ce.type === "Meta") {
            if (ce.subtype === "SetTempo") {
              tempo = createTempo(tempo.ppq, ce.data[0]);
            }
            if (ce.subtype === "TimeSignature") {
              sig.noteBeats = ce.data[0];
              sig.perMeasure = ce.data[1];
              sig.ticksInMetronome = ce.data[2];
              sig.toQuarter32nd = ce.data[3];
            }
          }
          if (ce.type === "Channel") {
            if (ce.subtype === "ProgramChange") {
              track.instrument = useInstrument(system2, run, ce.data[0]);
            }
            if (ce.subtype === "Controller") {
              if (ce.data[0] === 7) {
                track.vol.gain.setValueAtTime(ce.data[1] / 127, time);
              }
              if (ce.data[0] === 10) {
                const amt = (ce.data[1] - 127 + 63.5) / 63.5;
                const amtScaled = Math.round(amt * 10) / 10;
                track.pan.positionX.setValueAtTime(amtScaled, time);
              }
            }
            if (ce.subtype === "NoteOff") {
              const fns = track.lastFn[ce.channel ?? 0];
              while (fns.length) {
                const soundOff = fns.pop();
                if (soundOff)
                  soundOff();
              }
            }
            if (ce.subtype === "NoteOn") {
              const mnote = ce.data[0] ?? 60;
              const note = noteByMidi(mnote);
              if (ce.channel === 9) {
                track.lastFn[ce.channel].push(usePercussion(system2, run, mnote)(note, system2.currentTime(), track.pan));
              } else {
                if (track.instrument) {
                  track.vol.gain.setValueAtTime(ce.data[1] / 127, system2.currentTime());
                  track.lastFn[ce.channel ?? 0].push(track.instrument(note, system2.currentTime(), track.pan));
                } else {
                  console.warn(`NoteOn event, but no instrument selected. Channel ${ce.channel ?? 0} Track ${t} Event ${e}`);
                }
              }
            }
            if (ce.subtype === "EndOfTrack") {
              stopAllSounds(track);
            }
          }
        }
      }
      return [tempo, sig];
    };
    return tickCallBack;
  }
  function insertRun(system2, run) {
    let tempo = createTempo(run.song.ticksPerQuarterNote, 5e5);
    let sig = toSignature("4/4");
    run.tickListeners.push(handleEventCallback(system2, run));
    createPlayHeadClock(system2, tempo, sig, run);
  }
})();
//# sourceMappingURL=andy.js.map
